"""
Documentar.
"""

from typing import Any, Union, List, Dict


DEFAULT_TTL = 300
MISSING_KEY = object()


class BaseCache:

    def __init__(self, **params: Dict[str, Any]) -> None:
        self.default_ttl = int(str(params.get("ttl", DEFAULT_TTL)))

    def get_ttl(self, ttl: Union[int, None] = None) -> int:
        return ttl if (isinstance(ttl, int) and ttl >= 0) or (ttl is not None) else self.default_ttl

    def key_exists(self, key: str) -> bool:
        return self.get(key, MISSING_KEY) is not MISSING_KEY

    def add(self, key: str, value: Any, ttl: Union[int, None] = None) -> bool:
        """ Define um valor no cache caso a chave ainda não exista.

            Caso já exista, não define o valor no cache, preservando o TTL.

            Args:
                key (_type_): A chave a ser usada para identificar o valor no cache.
                value (_type_): O valor a ser posto em cache.
                ttl (_type_, optional): O tempo máximo que o item pode existir no cache. O padrão é None.

            Raises:
                NotImplementedError: _description_

            Returns:
                bool: True adicionou, ou seja, se ainda não existia. False se já existia.
        """
        raise NotImplementedError("subclasses of BaseCache must provide an add() method")

    def get(self, key: str, default: Any = None) -> Any:
        """ Busca uma determinada chave no cache. Se a chave não existir, retorne o padrão, que por padrão é None.
            Args:
                key (_type_): A chave a ser usada para identificar o valor no cache.
                default (_type_): O valor padrão, caso não seja encontrado no cache. O padrão é None.
            Raises:
                NotImplementedError: _description_
        """
        raise NotImplementedError("subclasses of BaseCache must provide a get() method")

    def set(self, key: str, value: Any, ttl: Union[int, None] = None) -> None:
        """ Set a value in the cache. If ttl is given, use that ttl for the key; otherwise use the default cache ttl.
        """
        raise NotImplementedError("subclasses of BaseCache must provide a set() method")

    def get_or_set(self, key: str, default: Any, ttl: Union[int, None] = None) -> Any:
        """ Fetch a given key from the cache. If the key does not exist,
            set the key and set it to the default value. The default value can
            also be any callable. If ttl is given, use that ttl for the
            key; otherwise use the default cache ttl.

            Return the value of the key stored or retrieved.
        """
        val = self.get(key)
        if val is None:
            self.set(key, default, ttl=ttl)
            val = default
        return val

    def delete(self, key: str) -> None:
        """ Delete a key from the cache and return whether it succeeded, failing silently.
        """
        raise NotImplementedError("subclasses of BaseCache must provide a delete() method")

    def touch(self, key: str, ttl: Union[int, None] = None) -> bool:
        """
            Update the key's expiry time using ttl. Return True if successful or False if the key does not exist.
        """
        raise NotImplementedError("subclasses of BaseCache must provide a touch() method")

    def get_many(self, keys: List[str]) -> Dict[str, Any]:
        """ Fetch a bunch of keys from the cache. For certain backends (memcached, pgsql) this can be *much* faster
            when fetching multiple values.

            Return a dict mapping each key in keys to its value. If the given key is missing, it will be missing from
            the response dict.
        """
        d = {}
        for k in keys:
            val = self.get(k, MISSING_KEY)
            if val is not MISSING_KEY:
                d[k] = val
        return d

    def set_many(self, data: Any, ttl: Union[int, None] = None) -> None:
        """ Set a bunch of values in the cache at once from a dict of key/value
            pairs.  For certain backends (memcached), this is much more efficient
            than calling set() multiple times.

            If ttl is given, use that ttl for the key; otherwise use the
            default cache ttl.
        """
        for key, value in data.items():
            self.set(key, value, ttl=ttl)

    def delete_many(self, keys: List[str]) -> None:
        """ Delete a bunch of values in the cache at once. For certain backends
            (memcached), this is much more efficient than calling delete() multiple times.
        """
        for key in keys:
            self.delete(key)

    def incr(self, key: str, delta: Union[int, float] = 1) -> Union[int, float]:
        """ NOT SAFE FOR CONCURRENCY. Add delta to value in the cache.
            If the key does not exist, raise a ValueError exception.
        """
        value = self.get(key, None)
        if value is None:
            raise ValueError("Key '%s' not found" % key)
        new_value = value + delta
        self.set(key, new_value)
        return new_value

    def decr(self, key: str, delta: Union[int, float] = 1) -> Union[int, float]:
        """ NOT SAFE FOR CONCURRENCY. Subtract delta from value in the cache.
            If the key does not exist, raise a ValueError exception.
        """
        return self.incr(key, -delta)

    def clear(self) -> None:
        """ Remove *all* values from the cache at once. """
        raise NotImplementedError("subclasses of BaseCache must provide a clear() method")

    def close(self, **kwargs: Dict[str, Any]) -> None:
        """ Close the cache connection """
        pass



from typing import Union, Dict, Any
import inspect
from rds_core.transformers.fields import Field


class ModelTransformer:
    __fields = None

    @classmethod
    def transform_to_dict(
        cls, source_dict: Union[Dict, None]
    ) -> Union[Dict[str, Any], None]:
        destination_dict = {}
        for destination_fieldname, transformer_field in cls.get_fields().items():
            transformer_field.transform_to_dict(
                source_dict, destination_dict, destination_fieldname
            )
        return destination_dict

    @classmethod
    def get_fields(cls) -> dict:
        if cls.__fields is None:
            cls.__fields = {}
            for k, v in inspect.getmembers(cls):
                if isinstance(v, Field):
                    cls.__fields[k] = v
        return cls.__fields



"""
Documentar.
"""
import dynaconf
from rds_core.helpers import env

environment = env("ENVIRONMENT", "local")

settings = dynaconf.Dynaconf(
    envvar_prefix="DYNACONF",
    settings_files=[f"settings.{environment}.{ext}" for ext in ["yaml", "toml"]],
)



"""
Documentar.
"""

import dynaconf
from rds_core.config import settings
from rds_core.helpers import instantiate_class


caches = {}

if 'CACHES' in settings and isinstance(settings.CACHES, dynaconf.utils.boxing.DynaBox):
    for cache_name, cache_configs in settings.CACHES.items():
        caches[cache_name] = instantiate_class(cache_configs['BACKEND'], **cache_configs.get('OPTIONS', {}))
    default_cache = caches.get('default', instantiate_class('rds_core.cache.nocache.NoCache'))
else:
    default_cache = instantiate_class('rds_core.cache.nocache.NoCache')



"""
Documentar.
"""

import datetime
from typing import Dict, Union, Any, List
import opensearchpy
import elasticsearch
from dynaconf.utils.boxing import DynaBox
from rds_core.config import settings
from rds_core.helpers import instantiate_class


__search_engine_cache: Dict[str, Any] = {}


def get_search_engine_config(alias: str = "default") -> DynaBox:
    return settings.SEARCH_ENGINES[alias]


def search_engine(
    alias: str = "default",
    username: Union[str, None] = None,
    password: Union[str, None] = None,
) -> Union[opensearchpy.OpenSearch, elasticsearch.Elasticsearch]:
    se_config = get_search_engine_config(alias)

    engine = se_config["engine"]
    params = {
        k: v
        for k, v in se_config.items()
        if k not in ("username", "password", "engine", "dsl_engine", "hosts")
    }
    username = username if username else se_config["username"]
    password = password if password else se_config["password"]
    params["http_auth"] = (username, password)
    client = instantiate_class(engine, se_config["hosts"].split(","), **params)
    __search_engine_cache[alias] = client
    return client


# def dsl_connection(alias: str = 'default') -> Union[opensearchpy.OpenSearch, elasticsearch.Elasticsearch]:
#     se_config = get_search_engine_config(alias)
#
#     dsl_connections = get_variable_by_pathname(se_config['dsl_engine'])
#     if alias not in dsl_connections.connections:
#         dsl_connections.add_connection(alias, search_engine(alias))
#     # dsl_connections.create_connection(
#     #     alias=alias,
#     #     http_auth=(se_config['user'], se_config['password']),
#     #     **{k:v for k,v in se_config.items() if k not in ('user', 'password', 'engine', 'dsl_engine')},
#     # )
#     return dsl_connections.get_connection(alias)


def create_index_if_not_exists(
    index: str,
    body: Union[dict, None] = None,
    params: Union[Dict[str, Any], None] = None,
    headers: Union[Dict[str, Any], None] = None,
    alias: str = "default"
) -> bool:
    """Cria um índice caso não exista. Retorna True se criou e False se não criou.

    Args:
        index (str): nome do índice a ser criado
        body (dict|None): corpo do índice
        params (dict|None): parametros
        headers (dict|None): cabeçalhos
        alias (str): alias para o search engine

    Raises:
        exception: Erro conforme retornado pelo Search Engine.

    Returns:
        bool: True se criou. False se não criou.
    """
    if not body:
        body = {}
    try:
        search_engine(alias).indices.create(index, body=body, params=params, headers=headers)  # type: ignore
        return True
    except Exception as e:
        if getattr(e, "error", None) == "resource_already_exists_exception":
            return False
        raise e


def delete_index_if_exists(index_name: str,
                           params: Union[Dict[str, Any], None] = None,
                           headers: Union[Dict[str, Any], None] = None,
                           alias: str = "default",
                           fail: bool = False) -> bool:
    """Apaga um índice caso exista. Retorna True se apagou e False se não apagou.

    Args:
        index_name (str): nome do índice a ser criado
        params (Dict[str, Any] | None):
        headers (Dict[str, Any] | None): Union[Dict[str, Any], None] = None,
        alias (str): alias para o search engine

    Raises:
        Exception: Erro conforme retornado pelo Search Engine.

    Returns:
        bool: True se criou. False se não criou.
    """
    try:
        search_engine(alias).indices.delete(index_name, params=params, headers=headers)  # type: ignore
        return True
    except Exception as e:
        if getattr(e, "error", None) == "index_not_found_exception":
            return False
        raise Exception(e)


def query(
    index_name: str,
    query_string: Union[str, int, float, datetime.date, datetime.datetime],
    fields: Union[List, None] = None,
    alias: str = "default",
) -> Any:
    if fields is None:
        fields = []
    response = search_engine(alias).search(
        index=index_name,
        body={
            "size": 5,
            "query": {"multi_match": {"query": query_string, "fields": fields}},
        },
    )
    return response["hits"]["hits"], response["hits"]["total"]["value"]


def search(index_name: str, body: dict, alias: str = "default") -> Any:
    response = search_engine(alias).search(index=index_name, body=body)
    return response["hits"]["hits"], response["hits"]["total"]["value"]


def index(
    index_name: str,
    body: Union[dict, None] = None,
    id: Any = None,
    params: Union[Dict[str, Any], None] = None,
    headers: Union[Dict[str, Any], None] = None,
    alias: str = "default"
) -> Union[Any, Any]:
    if not body:
        body = {}
    return search_engine(alias).index(index=index_name, body=body, id=id, params=params, headers=headers)  # type: ignore



def search_engine_healthy(params: Union[Dict[str, Any], None] = None,
                          headers: Union[Dict[str, Any], None] = None,
                          alias: str = "default") -> bool:
    print(search_engine(alias).ping)
    return search_engine(alias).ping(params=params, headers=headers)



"""
Documentar.
"""
from typing import Any, List, Dict
import logging
import json
from os import getenv
import importlib


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def str2bool(v):
    TRUE_STRS = ("true", "verdade", "yes", "sim", "t", "v", "y", "s", "1")
    FALSE_STRS = ("false", "falso", "no", "nao", "não", "f", "n", "0")

    if isinstance(v, bool):
        return v

    if v is None or (isinstance(v, str) and v.strip() == ""):
        return None

    if isinstance(v, int) and v in (1, 0):
        return v == 1

    if isinstance(v, str) and v.strip().lower() in TRUE_STRS + FALSE_STRS:
        return v.lower() in TRUE_STRS

    raise ValueError("Boolean value expected.")


def env(name, default=None, wrapped=False):
    result = getenv(name, default)
    if wrapped and isinstance(result, str) and result[0:1] == "'" and result[-1:] == "'":
        return result[1:-1]
    return result


def env_as_list(name, default="", delimiter=",", wrapped=False):
    result = env(name, default, wrapped)
    if result is None:
        return None
    if type(result) == str:
        if result.strip() == "" and default.strip() == "":
            return []
        return result.split(delimiter)
    if type(result) in (list, tuple):
        return list(result)
    raise TypeError("env_as_list requires str, list or tuple as default")


def env_as_list_of_maps(name, key, default="", delimiter=",", wrapped=False):
    return [{key: x} for x in env_as_list(name, default, delimiter, wrapped)]


def env_as_bool(name, default=None, wrapped=False):
    return str2bool(env(name, default, wrapped))


def env_from_json(key, default="", wrapped=False):
    result = env(key, default, wrapped)
    return json.loads(result) if result is not None else result


def env_as_int(key, default=None, wrapped=False):
    result = env(key, default, wrapped)
    return int(result) if result is not None else result


def get_class(full_class_name: str) -> Any:
    module_name, class_name = full_class_name.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), class_name)


def instantiate_class(full_class_name: str, *args: List, **kwargs: Dict[str, Any]) -> Any:
    Klass = get_class(full_class_name)
    return Klass(*args, **kwargs)


def get_variable_by_pathname(full_class_name: str) -> Any:
    module_name, class_name = full_class_name.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), class_name)


def get_dict_by_pathname(obj: dict, ref: str) -> Any:
    """
    Use MongoDB style 'something.by.dot' syntax to retrieve objects from Python dicts.

    Usage:
       >>> x = {"top": {"middle" : {"nested": "value"}}}
       >>> q = 'top.middle.nested'
       >>> get_dict_by_pathname(x,q)
       "value"

    Credit: https://gist.github.com/mittenchops/5664038
    """
    val = obj
    tmp = ref
    ref = tmp.replace(".XX", "[0]")
    if tmp != ref:
        logger.warning("Warning: replaced '.XX' with [0]-th index")
    for key in ref.split("."):
        idstart = key.find("[")
        embedslist = 1 if idstart > 0 else 0
        if embedslist:
            idx = int(key[idstart + 1 : key.find("]")])
            kyx = key[:idstart]
            try:
                val = val[kyx][idx]
            except IndexError:
                logger.warning(f"Index: x['{kyx}'][{idx}] does not exist.")
                raise
        else:
            val = val.get(key, None) if val is not None else None
    return val



import logging
from typing import Union, Dict, Any, List
from datetime import datetime
from rds_core.helpers import get_dict_by_pathname

strptime = datetime.strptime


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Field:
    def __init__(self, source_field: Union[str, object, None]):
        self.source_field = source_field

    def transform_to_dict(
        self,
        source_dict: Dict[str, Any],
        dest_dict: Dict[str, Any],
        dest_fieldname: str,
    ):
        if self.source_field is None:
            dest_dict[dest_fieldname] = None
        else:
            dest_dict[dest_fieldname] = self.cast_source_value(
                get_dict_by_pathname(source_dict, self.source_field)
            )

    def cast_source_value(self, source_value: Union[str, int, bool, float]):
        return source_value


class CharField(Field):
    pass


class DateField(Field):
    def __init__(self, source_field: str, format: str = "%d/%m/%Y"):
        super().__init__(source_field)
        self.format = format

    def cast_source_value(
        self, source_value: Union[None, str, int, bool, float]
    ) -> datetime.date:
        return strptime(source_value, self.format) if source_value is not None else None


class DateTimeField(Field):
    def __init__(self, source_field: str, format: str = "%d/%m/%Y %H:%M:%S"):
        super().__init__(source_field)
        self.format = format

    def cast_source_value(self, source_value: Union[str, int, bool, float]) -> datetime:
        return strptime(source_value, self.format) if source_value is not None else None


class SimpleConcatField(Field):
    def __init__(self, fields: Union[List[str], None] = None, separator: str = ""):
        super().__init__(None)
        self.fields = fields
        self.separator = separator

    def transform_to_dict(
        self,
        source_dict: Dict[str, Any],
        dest_dict: Dict[str, Any],
        dest_fieldname: str,
    ) -> str:
        if self.fields is not None:
            dest_dict[dest_fieldname] = self.separator.join(
                [
                    self.cast_source_value(get_dict_by_pathname(source_dict, f))
                    for f in self.fields
                ]
            )


class SubModelField(Field):
    def transform_to_dict(
        self,
        source_dict: Dict[str, Any],
        dest_dict: Dict[str, Any],
        dest_fieldname: str,
    ):
        dest_dict[dest_fieldname] = self.source_field.transform_to_dict(source_dict)



"""
Documentar.
"""

from typing import Any, List, Dict, Union

import opensearchpy
import elasticsearch
from rds_core.cache.base import BaseCache
from rds_core.searchengine import search_engine as search_engine_func


class SearchEngineCache(BaseCache):

    def __init__(self, **params: Dict[str, Any]) -> None:
        super(SearchEngineCache, self).__init__(**params)
        self.search_engine_alias: str = str(params.get("search_engine_alias", 'default'))
        self.index_name: str = str(params.get("index_name", 'cnes_establecimento_cache'))
        self.refresh: str = str(params.get("refresh", 'wait_for'))

    @property
    def search_engine(self) -> Any:
        return search_engine_func(self.search_engine_alias)

    def key_exists(self, key: str) -> bool:
        return self.search_engine.exists(self.index_name, id=key)

    def add(self, key: str, value: Any, ttl: Union[int, None] = None) -> bool:
        """ Define um valor no cache caso a chave ainda não exista.

            Caso já exista, não define o valor no cache, preservando o TTL.

            Args:
                key (_type_): A chave a ser usada para identificar o valor no cache.
                value (_type_): O valor a ser posto em cache.
                ttl (_type_, optional): O tempo máximo que o item pode existir no cache. O padrão é None.

            Raises:
                NotImplementedError: _description_

            Returns:
                bool: True adicionou, ou seja, se ainda não existia. False se já existia.
        """
        if self.key_exists(key):
            return False
        body = {'value': value, 'ttl': self.get_ttl(ttl)}
        self.search_engine.create(self.index_name, body=body, id=key, refresh=self.refresh)
        return True

    def get(self, key: str, default: Any = None) -> Any:
        try:
            response = self.search_engine.get(self.index_name, key)
            doc = response['_source']
            # ttl = doc['ttl']
            return doc['value'] if doc['value'] is not None else default
        except opensearchpy.exceptions.NotFoundError:
            return default
        except elasticsearch.exceptions.NotFoundError:
            return default

    def set(self, key: str, value: Any, ttl: Union[int, None] = None) -> None:
        if type(value) == bytes:
            raise ValueError("Não suporta salvar bytes")
        body = {'value': value, 'ttl': self.get_ttl(ttl)}
        self.search_engine.index(self.index_name, body=body, id=key, refresh=self.refresh)

    def touch(self, key: str, ttl: Union[int, None] = None) -> bool:
        if self.key_exists(key):
            self.search_engine.index(self.index_name, body={'ttl': self.get_ttl(ttl)}, id=key, refresh=self.refresh)
            return True
        return False

    def delete(self, key: str) -> None:
        try:
            self.search_engine.delete(self.index_name, id=key, refresh=self.refresh)
        except opensearchpy.exceptions.NotFoundError:
            return None
        except elasticsearch.exceptions.NotFoundError:
            return None

    def clear(self) -> None:
        raise NotImplementedError("subclasses of BaseCache must provide a clear() method")

    def get_many(self, keys: List[str]) -> Dict[str, Any]:
        """ Fetch a bunch of keys from the cache. For certain backends (memcached, pgsql) this can be *much* faster
            when fetching multiple values.

            Return a dict mapping each key in keys to its value. If the given key is missing, it will be missing from
            the response dict.
        """
        response = self.search_engine.mget({'ids': [k for k in keys]}, index=self.index_name)
        return {x['_id']: x['_source']['value'] for x in response['docs'] if x['found']}



import os
from setuptools import setup


def fast_scandir(dirname):
    subfolders = [f.path for f in os.scandir(dirname) if f.is_dir()]
    for dirname in list(subfolders):
        subfolders.extend(fast_scandir(dirname))
    return subfolders


def package_data_dirs(root, data_dirs):
    data_dirs_path = [x + "/*" for x in data_dirs]
    for data_dir in data_dirs:
        data_dirs_path += [
            x.replace(f"{root}/", "") + "/*" for x in fast_scandir(f"{root}/{data_dir}")
        ]

    return {root: data_dirs_path}


requirements = [
    # config
    "dynaconf>=3.1.11",
    # HTTP clients
    "requests>=2.28.1",
    # LAIS public Python libraries
    "dbc-reader>=0.1.1",
    # Search engine
    "opensearch_py>=2.0.0",
    "opensearch_dsl>=2.0.1",
    "elasticsearch>=7.17.6",
    "elasticsearch_dsl>=7.4.0",
]

with open("requirements.txt", "w") as file1:
    for requirement in requirements:
        file1.write(f"{requirement}\n")

setup(
    name="rds-core",
    version="0.1.1",
    description="Framework para serviços do Rede de Dados em Saúde do LAIS",
    long_description="É uma biblioteca pública em Python que condensa um conjunto de boas práticas para o"
    " desenvolvimento das aplicações que compõem a Rede da Dados em Saúde (RDS) RDS do Laboratório de"
    " Inovação Tecnológica em Saúde (LAIS) e dos parceiros que contam com o LAIS para fazer suas próprias"
    " RDS, a exemplo RDS-RN e RDS-ES.",
    author="Kelson da Costa Medeiros",
    author_email="kelson.medeiros@lais.huol.ufrn.br",
    keywords=["rds", "cache", "config", "helper", "searchengine"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires=">=3.9",
    install_requires=requirements,
    packages=[
        "rds_core",
        "rds_core.cache",
        "rds_core.config",
        "rds_core.helpers",
        "rds_core.searchengine",
        "rds_core.transformers",
        "rds_core.transformers.fields",
        "rds_core.transformers.models",
    ],
    package_dir={"rds_core": "rds_core"},
    package_data=package_data_dirs("rds_core", []),
    download_url="https://github.com/lais-huol/rds-core/tags",
    url="https://github.com/lais-huol/rds-core",
)



"""
Documentar.
"""

from typing import Any, List, Dict, Union
from rds_core.cache.base import BaseCache


class NoCache(BaseCache):
    def __init__(self, *args: List, **kwargs: Dict[str, Any]) -> None:
        super().__init__(*args, **kwargs)

    def add(self, key: str, value: Any, ttl: Union[int, None] = None) -> bool:
        return True

    def get(self, key: str, default: Any = None) -> Any:
        return default

    def set(self, key: str, value: Any, ttl: Union[int, None] = None) -> None:
        pass

    def delete(self, key: str) -> None:
        return

    def clear(self) -> None:
        return



"""
RDS é **acrônimo** para Rede de Dados em Saúde. Este framework/api condensa um conjunto de boas práticas para o
desenvolvimento das aplicações que compõem ou comporão a RDS do LAIS e dos parceiros que contratarem o LAIS para fazer
suas próprias RDS, a exemplo REDS-RN e RDS-ES.
"""



"""
Documentar.
"""


def is_cpf_or_cns(cpf_or_cns: str) -> str:
    if len(cpf_or_cns) == 15:
        return 'cns'
    elif len(cpf_or_cns) == 11:
        return 'cpf'
    else:
        raise Exception('Deve ter 15 (cns) ou 11 (cpf) digitos')



"""
Documentar.
"""

from typing import Dict, List, Any, Union
import json
from http.client import HTTPException as OriginalHTTPException
import requests
from requests.structures import CaseInsensitiveDict
from rds_core.config import settings

DEFAULT_HEADERS: Dict[str, Any] = settings.get('DEFAULT_HEADERS', {})


class HTTPException(OriginalHTTPException):
    def __init__(self,
                 message: str,
                 url: str,
                 status_code: Union[str, None] = None,
                 reason: Union[str, None] = None,
                 request_headers: Union[Dict[Any, Any], None] = None,
                 response_headers: Union[CaseInsensitiveDict, None] = None) -> None:
        super().__init__(message)
        self.url = url
        self.status_code = status_code
        self.reason = reason
        self.request_headers = request_headers
        self.response_headers = response_headers


def get(url: str, headers: Dict[str, str] = {}, encoding: str = 'utf-8', decode: bool = True, **kwargs) -> Any:

    _headers = {**DEFAULT_HEADERS, **headers}
    response = requests.get(url, headers=_headers, **kwargs)

    if response.ok:
        byte_array_content = response.content
        return byte_array_content.decode(encoding) if decode and encoding is not None else byte_array_content
    else:
        message = f'{response.status_code} - {response.reason}'
        raise HTTPException(message, url, str(response.status_code), response.reason, _headers, response.headers)


def get_json(url: str, headers: Dict[str, str] = {}, encoding: str = 'utf-8',
             json_kwargs: Dict[str, Any] = {}, **kwargs) -> Union[List[Any], Dict[str, Any]]:
    content = get(url, headers=headers, encoding=encoding, **kwargs)
    return json.loads(content, **json_kwargs)
